<?php

declare(strict_types=1);
/**
 * This file is part of Hyperf.
 *
 * @link     https://www.hyperf.io
 * @document https://doc.hyperf.io
 * @contact  group@hyperf.io
 * @license  https://github.com/hyperf/hyperf/blob/master/LICENSE
 */

namespace HyperfTest\JsonRpc;

use Hyperf\Di\Container;
use Hyperf\JsonRpc\JsonRpcPoolTransporter;
use Hyperf\JsonRpc\Packer\JsonLengthPacker;
use Hyperf\JsonRpc\Pool\Frequency;
use Hyperf\JsonRpc\Pool\PoolFactory;
use Hyperf\JsonRpc\Pool\RpcPool;
use Hyperf\LoadBalancer\Node;
use Hyperf\Pool\Channel;
use Hyperf\Pool\PoolOption;
use Hyperf\Utils\ApplicationContext;
use HyperfTest\JsonRpc\Stub\RpcPoolStub;
use Mockery;
use PHPUnit\Framework\TestCase;

/**
 * @internal
 * @coversNothing
 */
class JsonRpcPoolTransporterTest extends TestCase
{
    protected function tearDown()
    {
        Mockery::close();
    }

    public function testJsonRpcPoolTransporterConfig()
    {
        $factory = Mockery::mock(PoolFactory::class);
        $transporter = new JsonRpcPoolTransporter($factory, [
            'pool' => ['min_connections' => 10],
            'settings' => $settings = [
                'open_length_check' => true,
                'package_length_type' => 'N',
                'package_length_offset' => 0,
                'package_body_offset' => 4,
            ],
        ]);

        $this->assertSame(10, $transporter->getConfig()['pool']['min_connections']);
        $this->assertSame(32, $transporter->getConfig()['pool']['max_connections']);
        $this->assertSame($settings, $transporter->getConfig()['settings']);
    }

    public function testJsonRpcPoolTransporterSendLengthCheck()
    {
        $container = $this->getContainer();
        $factory = $container->get(PoolFactory::class);
        $transporter = new JsonRpcPoolTransporter($factory, $options = [
            'pool' => ['min_connections' => 10],
            'settings' => [
                'open_length_check' => true,
                'package_length_type' => 'N',
                'package_length_offset' => 0,
                'package_body_offset' => 4,
            ],
        ]);
        $transporter->setNodes([new Node('127.0.0.1', 9504)]);

        $packer = new JsonLengthPacker($options);

        $string = $transporter->send($packer->pack($data = ['id' => $id = uniqid()]));

        $this->assertSame($data, $packer->unpack($string));
    }

    public function testJsonRpcPoolTransporterSendEofCheck()
    {
        $container = $this->getContainer();
        $factory = $container->get(PoolFactory::class);
        $transporter = new JsonRpcPoolTransporter($factory, $options = [
            'pool' => ['min_connections' => 10],
            'settings' => [
                'open_eof_split' => true,
                'package_eof' => "\r\n",
            ],
        ]);
        $transporter->setNodes([new Node('127.0.0.1', 9504)]);

        $packer = new JsonLengthPacker($options);

        $string = $transporter->send($packer->pack($data = ['id' => $id = uniqid()]));

        $this->assertSame($data, $packer->unpack($string));
    }

    public function testsplObjectHash()
    {
        $class = new \stdClass();
        $class->id = 1;
        $hash = spl_object_hash($class);

        $class->id = 2;
        $hash2 = spl_object_hash($class);

        $this->assertSame($hash, $hash2);
    }

    protected function getContainer()
    {
        $container = Mockery::mock(Container::class);
        ApplicationContext::setContainer($container);

        $container->shouldReceive('get')->with(PoolFactory::class)->andReturn(new PoolFactory($container));
        $container->shouldReceive('make')->with(RpcPool::class, Mockery::any())->andReturnUsing(function ($_, $args) use ($container) {
            return new RpcPoolStub($container, $args['name'], $args['config']);
        });
        $container->shouldReceive('make')->with(Frequency::class, Mockery::any())->andReturn(null);
        $container->shouldReceive('make')->with(PoolOption::class, Mockery::any())->andReturnUsing(function ($_, $args) {
            return new PoolOption(...array_values($args));
        });
        $container->shouldReceive('make')->with(Channel::class, Mockery::any())->andReturnUsing(function ($_, $args) {
            return new Channel(10);
        });

        return $container;
    }
}
